unit Cd40u1;
(*Simple substitution code solver.
Only changes 'a's (ord('a')=97) to 'z's (122), and
all encoding characters must also come from a..z
File will have a two byte checksum at it's end, consisting
of (the sum of the plaintext characters)mod 256,
experessed in two hex digits. I.e. if the sum were
32, the checksum digits would be 10, bytes 49 and 48.
(CD37 encodes text into such files)
Other characters may be present (spaces, etc), but they won't
be encoded, hence should not be decoded. There should be
nothing from A..Z (the capital forms), but if there is
they, too, should be unencoded.

Program does not ensure that last two bytes are in fact a meaningful checksum.

NOTES FOR THOSE WHO WOULD EXTEND THIS PROGRAM:

Once buOpenFileClick has been executed...

The encoded text is available in the string array sToDecode,
    (as well as in the stringgrid)
If you alter what is in the array bDecodeTable, and then
    call UpdateDisplay, that is all you need to do to
    refresh the display with your latest attempt at a decode.
The test for whether the decode meets the requirements of
    the checksum is within the UpdateDisplay procedure.
*)

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Grids, StdCtrls, Menus;

type
  Tcd40f1 = class(TForm)
    sgText: TStringGrid;
    OpenDialog1: TOpenDialog;
    Label1: TLabel;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Quit1: TMenuItem;
    Help1: TMenuItem;
    meHelp: TMemo;
    buHelpOK: TButton;
    procedure FormCreate(Sender: TObject);
    procedure buQuitClick(Sender: TObject);
    procedure buOpenFileClick(Sender: TObject);
    procedure sgTextSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure sgTextSetEditText(Sender: TObject; ACol, ARow: Longint;
      const Value: String);
    procedure Help1Click(Sender: TObject);
    procedure buHelpOKClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  cd40f1: Tcd40f1;

implementation

{$R *.DFM}
{$R+}
const ver='10 Apr 00';
  MaxRows=45;
  MaxCols=50;(*N.B.The string grid will need 2*MaxRows rows*)

var bTmp,bGR,bGC,bValueWas,bRowWas,bColWas,
bCSToMatch:byte;
  (*bGR,GC:Rows and columns in grid. cells[0,0] is upper left,
     cells bGR,bGC is lower right
    bValueWas: Holds value that was in a cell when it was selected.
    bRowWas,bColWas.. associated with bValueWas*)
    sTmp:string;
    sToDecode:array[0..MaxRows] of string[MaxCols];
    bDecodeTable: array[0..25] of byte;
    wCharsInMsg,wDone,wCS:word;(*Count of chars in message, incl checksum
    digits, and chars dealt with while traversing message*)
    boFileIsOpen:boolean;

procedure UpdateDisplay;
(*This, and remembering how it works, is the heart
of making this program work, and remain extendable.

This proc goes through the "decoded" lines of the display,
putting in the decode of any letter for which a decode has
been assigned.

The decodes are held in the array bDecodeTable.
If bDecodeTable[0]:=0, then no decode has been assigned for
any 'a's in the encoded text. If, on the other hand,
bDecodeTable[0]:=98, then each 'a' in the encoded text
is thought to translate to 'b' (chr(98)='b')
*)
var bTmpCol,bTmpRow:byte;
begin
if boFileIsOpen then begin
bTmpRow:=0;
wDone:=0;
wCS:=0;
repeat
for bTmpCol:=0 to bGC do
  if wDone<wCharsInMsg-2 then begin (*The minus 2 is to skip Checksum chars*)
    inc(wDone);
    bTmp:=ord(CD40F1.sgText.cells[bTmpCol,bTmpRow][1]);
    if (bTmp>96)and (bTmp<123) then
    CD40F1.sgText.cells[bTmpCol,bTmpRow+1]:=chr(bDecodeTable[bTmp-97])(*no ; here*)
    else CD40F1.sgText.cells[bTmpCol,bTmpRow+1]:=chr(bTmp);
    wCS:=wCS+ord(CD40F1.sgText.cells[bTmpCol,bTmpRow+1][1]);
    wCS:=wCS and $FF;
    end;(*for bTMPCol..*)
bTmpRow:=bTmpRow+2;
until bTmpRow>bGR;
if bCSToMatch=wCS then sTmp:='Congratulations! Appears to be fully decoded' (*no ; here*)
else sTmp:='Checksum still shows message not fully decoded';
CD40F1.label1.caption:=sTmp;
end;(*if boFileIs..*)
end;

procedure Tcd40f1.FormCreate(Sender: TObject);
begin
application.title:='CD40';
label1.caption:='Welcome to CD40: Code solver';
meHelp.hide;
meHelp.align:=alclient;
buHelpOK.hide;
boFileIsOpen:=false;(*To prevent "update" before data present*)
cd40f1.caption:=cd40f1.caption+' ver: '+ver;
bGR:=sgText.rowcount-1;
bGC:=sgText.colcount-1;
(*a 3x3 grid has row/col count of 3, but everything else in prog calls last row/col "2"*)
bValueWas:=ord(sgText.cells[0,0][0]);
end;

procedure Tcd40f1.buQuitClick(Sender: TObject);
begin
application.terminate;
end;

function HexAsciiToDec(bTmp:byte):byte;
(*Simply returns 0 if value not 0..9, A..F, or a..f*)
begin
case bTmp of
48..57:HexAsciiToDec:=bTmp-48;
97..102:HexAsciiToDec:=bTmp-87;
65..70:HexAsciiToDec:=bTmp-55(*no ; here*)
else HexAsciiToDec:=0;
end;(*case*)
end;


procedure Tcd40f1.buOpenFileClick(Sender: TObject);
var dfin:file of byte;
  c1,bRow,bCol,bRowIndex:byte;
begin
  if OpenDialog1.Execute then begin
    (*clear decode table*)
    for c1:=0 to 25 do bDecodeTable[c1]:=0;
    (*read file, etc*)
    for bCol:=0 to bGC do
      for bRow:=0 to bGR do sgText.cells[bCol,bRow]:='';
    sTmp:=OpenDialog1.Files[0];
    assignfile(dfIn,sTmp);
    reset(dfIn);
    bRow:=0;bCol:=0;bRowIndex:=0;
    sToDecode[0]:='';
    wCharsInMsg:=0;
    repeat
      read(dfin,bTmp);
      inc(wCharsInMsg);
      sgText.cells[bCol,bRow]:=chr(bTmp);
      sToDecode[bRowIndex]:=sToDecode[bRowIndex]+chr(bTmp);
      inc(bCol);
      if bCol>bGC then begin
         bCol:=0;
         bRow:=bRow+2;
         (*Put bRow-too-big test in here eventually*)
         inc(bRowIndex);
         sToDecode[bRowIndex]:='';
         end;
    until eof(dfin);
    (*Pick up CS, convert to a byte*)
    seek(dfIn,wCharsInMsg-2);
    read(dfIn,bTmp);
    bCSToMatch:=HexAsciiToDec(bTmp)*16;
    read(dfIn,bTmp);
    bCSToMatch:=bCSToMatch+HexAsciiToDec(bTmp);
    closefile(dfin);
    bColWas:=sgText.col;
    bRowWas:=sgText.row;
    bValueWas:=ord(sgText.cells[bColWas,bRowWas][1]);
    boFileIsOpen:=true;
    end; (*OpenDialog1.Execute*)
end;

{Example

The following code appends 'My ' to any text
entered in StringGrid1.

procedure TForm1.StringGrid1GetEditText(Sender: TObject;
  ACol,  ARow: Longint; var Value: OpenString);
begin
  Value := 'My ' + Value;
end;

 }
procedure Tcd40f1.sgTextSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
begin
if sgText.cells[col,row]='' then bValueWas:=0(*no ; here*)
    else bValueWas:=ord(sgText.cells[col,row][1]);
bRowWas:=row;
bColWas:=col;
end;

procedure Tcd40f1.sgTextSetEditText(Sender: TObject; ACol, ARow: Longint;
  const Value: String);
begin
(*Next bit: Prevents "byhand" changes of values*)
if (ARow mod 2)=0 then
  if bValueWas=0 then sgText.cells[bColWas,bRowWas]:=''(*no ; here*)
    else sgText.cells[bColWas,bRowWas]:=chr(bValueWas);
(*Next deals with consequences of a change of a "Plaintext"row elelment*)
if (ARow mod 2)=1 then begin
  bTmp:=ord(sgText.cells[ACol,ARow-1][1]);
  if (value<>'') and (bTmp>96) and (bTmp<123) then begin(*do this only for a..z*)
    bDecodeTable[bTmp-97]:=ord(value[1]);
    UpdateDisplay;
    end;(*if bTmp>96..*)
  end;
end;

procedure Tcd40f1.Help1Click(Sender: TObject);
begin
meHelp.show;
buHelpOK.show;
end;

procedure Tcd40f1.buHelpOKClick(Sender: TObject);
begin
meHelp.hide;
buHelpOK.hide;
end;

end.
